from functools import partial
from typing import Any, Callable, Dict, Union

import os
import anndata as an
import numpy as np
import pandas as pd
import pytorch_lightning as pl
import torch
import torch.nn.functional as F
from pytorch_tabnet.tab_network import TabNet
from pytorch_tabnet.utils import create_explain_matrix
from scipy.sparse import csc_matrix
from torchmetrics.functional.classification.stat_scores import _stat_scores_update
from tqdm import tqdm
import torch.utils.data

from scsims.data import CollateLoader
from scsims.inference import DatasetForInference
from scsims.temperature_scaling import _ECELoss
from torchmetrics import Accuracy, F1Score, Precision, Recall, Specificity

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

class SIMSClassifier(pl.LightningModule):
    def __init__(
        self,
        input_dim,
        output_dim,
        n_d=8,
        n_a=8,
        n_steps=3,
        gamma=1.3,
        cat_idxs=[],
        cat_dims=[],
        cat_emb_dim=1,
        n_independent=2,
        n_shared=2,
        epsilon=1e-15,
        virtual_batch_size=128,
        momentum=0.02,
        mask_type="sparsemax",
        lambda_sparse=1e-3,
        optim_params: Dict[str, float] = None,
        scheduler_params: Dict[str, float] = None,
        weights: torch.Tensor = None,
        loss: Callable = None,  # will default to cross_entropy
        pretrained: bool = None,
        no_explain: bool = False,
        genes: list[str] = None,
        cells: list[str] = None,
        label_encoder: Callable = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__()
        self.save_hyperparameters()

        self.genes = genes
        self.cells = cells
        self.label_encoder = label_encoder

        # Stuff needed for training
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.lambda_sparse = lambda_sparse
        self.optim_params = optim_params
        self.weights = weights
        self.loss = loss

        if self.loss is None:
            self.loss = F.cross_entropy

        if pretrained is not None:
            self._from_pretrained(**pretrained.get_params())

        self.metrics = {
            "train": {x: y.to(device) for x, y in aggregate_metrics(num_classes=self.output_dim).items()},
            "val": {x: y.to(device) for x, y in aggregate_metrics(num_classes=self.output_dim).items()},
        }

        self.optim_params = (
            optim_params
            if optim_params is not None
            else {
                "optimizer": torch.optim.Adam,
                "lr": 3e-4,
                "weight_decay": 1e-8,
            }
        )

        self.scheduler_params = (
            scheduler_params
            if scheduler_params is not None
            else {
                "scheduler": torch.optim.lr_scheduler.ReduceLROnPlateau,
                "factor": 0.75,  # Reduce LR by 25% on plateau
            }
        )

        print(f"Initializing network")
        self.network = TabNet(
            input_dim=input_dim,
            output_dim=output_dim,
            n_d=n_d,
            n_a=n_a,
            n_steps=n_steps,
            gamma=gamma,
            cat_idxs=cat_idxs,
            cat_dims=cat_dims,
            cat_emb_dim=cat_emb_dim,
            n_independent=n_independent,
            n_shared=n_shared,
            epsilon=epsilon,
            virtual_batch_size=virtual_batch_size,
            momentum=momentum,
            mask_type=mask_type,
        )

        print(f"Initializing explain matrix")
        if not no_explain:
            self.reducing_matrix = create_explain_matrix(
                self.network.input_dim,
                self.network.cat_emb_dim,
                self.network.cat_idxs,
                self.network.post_embed_dim,
            )

        self._inference_device = "cuda:0" if torch.cuda.is_available() else "cpu"
        self.temperature = torch.nn.Parameter(torch.ones(1) * 1.5)

    def forward(self, x):
        logits, M_loss = self.network(x)
        # temp scaling will be 1 so logits wont change until model is calibrated
        return self.temperature_scale(logits), M_loss

    def _step(self, batch, tag):
        x, y = batch
        logits, M_loss = self.network(x)

        loss = self.loss(logits, y, weight=self.weights)
        loss = loss - self.lambda_sparse * M_loss

        # take softmax for metrics
        probs = logits.softmax(dim=-1)

        # if binary, probs will be (batch, 2), so take second column
        if probs.shape[-1] == 2:
            probs = probs[:, 1]

        tp, fp, _, fn = _stat_scores_update(
            preds=logits,
            target=y,
            num_classes=self.output_dim,
            reduce="macro",
        )

        return {
            "loss": loss,
            "tp": tp,
            "fp": fp,
            "fn": fn,
            "probs": probs,
        }

    # Calculations on step
    def training_step(self, batch, batch_idx):
        results = self._step(batch, "train")
        self.log(f"train_loss", results["loss"], on_epoch=True, on_step=True)
        for name, metric in self.metrics["train"].items():
            value = metric(results["probs"], batch[1])
            self.log(f"train_{name}", value=value)

        return results["loss"]

    def on_train_epoch_end(self) -> None:
        for name, metric in self.metrics["train"].items():
            value = metric.compute()
            self.log(f"train_{name}", value=value)
            metric.reset() # inplace

    def validation_step(self, batch, batch_idx):
        results = self._step(batch, "val")
        self.log(f"val_loss", results["loss"], on_epoch=True, on_step=True)
        for name, metric in self.metrics["val"].items():
            value = metric(results["probs"], batch[1])
            self.log(f"val_{name}", value=value)

        return results["loss"]

    def on_validation_epoch_end(self) -> None:
        for name, metric in self.metrics["val"].items():
            value = metric.compute()
            self.log(f"val_{name}", value=value)
            metric.reset()
    
    def configure_optimizers(self):
        if "optimizer" in self.optim_params:
            optimizer = self.optim_params.pop("optimizer")
            optimizer = optimizer(self.parameters(), **self.optim_params)
        else:
            optimizer = torch.optim.Adam(self.parameters(), **self.optim_params)
        print(f"Initializing with {optimizer = }")

        if self.scheduler_params is not None:
            scheduler = self.scheduler_params.pop("scheduler")
            scheduler = scheduler(optimizer, **self.scheduler_params)
            print(f"Initializating with {scheduler = }")

        if self.scheduler_params is None:
            return optimizer

        return {
            "optimizer": optimizer,
            "lr_scheduler": scheduler,
            "monitor": "train_loss",
        }

    def _parse_data(
            self,
            inference_data,
            batch_size=64,
            num_workers=os.cpu_count(),
            rows=None,
            **kwargs
        ) -> torch.utils.data.DataLoader:

        if isinstance(inference_data, str):
            inference_data = an.read_h5ad(inference_data)
        
        # handle zero inflation or deletion
        inference_genes = list(inference_data.var_names)
        training_genes = list(self.genes)

        # more inference genes than training genes
        assert len(set(inference_genes).intersection(set(training_genes))) > 0, "inference data shares zero genes with training data, double check the string formats and gene names"
        if len(inference_genes) - len(training_genes) > 0:
            inference_data = inference_data[:, training_genes].copy()
        else:
            diff = list(set(training_genes) - set(inference_genes))
            print(f"Inference data has {len(diff)} less genes than training; performing zero inflation.")

            zero_inflation = an.AnnData(X=np.zeros((inference_data.shape[0], len(diff))), obs=inference_data.obs)
            zero_inflation.var_names = diff
            inference_data = an.concat([zero_inflation, inference_data], axis=1)

        # now make sure the columns are the correct order
        inference_data = inference_data[:, training_genes].copy()

        if isinstance(inference_data, an.AnnData):
            inference_data = DatasetForInference(inference_data.X[rows, :] if rows is not None else inference_data.X)

        if not isinstance(inference_data, torch.utils.data.DataLoader):
            inference_data = CollateLoader(
                dataset=inference_data,
                batch_size=batch_size,
                num_workers=num_workers,
                **kwargs,
            )

        return inference_data

    def explain(
        self,
        anndata,
        rows=None,
        batch_size=64,
        num_workers=os.cpu_count(),
        currgenes=None,
        refgenes=None,
        normalize=False,
        **kwargs,
    ) -> tuple[np.ndarray, np.ndarray]:
        loader = self._parse_data(anndata, batch_size=batch_size, num_workers=num_workers, rows=rows, currgenes=currgenes, refgenes=refgenes, **kwargs)

        self.network.eval()
        res_explain = []

        all_labels = np.empty(len(loader.dataset))
        all_labels[:] = np.nan

        for batch_nb, data in enumerate(tqdm(loader)):
            print("DATA SHAPE IS", data.shape)
            # if we are running this on already labeled pairs and not just for inference
            if isinstance(data, tuple):
                X, label = data
                all_labels[batch_nb * batch_size : (batch_nb + 1) * batch_size] = label
            else:
                X = data

            M_explain, masks = self.network.forward_masks(X)
            for key, value in masks.items():
                masks[key] = csc_matrix.dot(
                    value.cpu().detach().numpy(), self.reducing_matrix
                )

            original_feat_explain = csc_matrix.dot(M_explain.cpu().detach().numpy(),
                                                   self.reducing_matrix)
            res_explain.append(original_feat_explain)

            if batch_nb == 0:
                res_masks = masks
            else:
                for key, value in masks.items():
                    res_masks[key] = np.vstack([res_masks[key], value])

        res_explain = np.vstack(res_explain)

        if normalize:
            res_explain /= np.sum(res_explain, axis=1)[:, None]

        return res_explain, all_labels

    def _compute_feature_importances(self, dataloader):
        M_explain, _ = self.explain(dataloader, normalize=False)
        sum_explain = M_explain.sum(axis=0)
        feature_importances_ = sum_explain / np.sum(sum_explain)

        return feature_importances_

    def feature_importances(self, dataloader, cache=False):
        if cache and self._feature_importances is not None:
            return self._feature_importances
        else:
            f = self._compute_feature_importances(dataloader)
            if cache:
                self._feature_importances = f
            return f

    def predict(self, inference_data: Union[str, an.AnnData, np.array], batch_size=32, num_workers=4, rows=None, currgenes=None, refgenes=None, **kwargs):
        print("parsing data...")
        loader = self._parse_data(
            inference_data,
            batch_size=batch_size,
            num_workers=num_workers,
            rows=rows,
            currgenes=currgenes,
            refgenes=refgenes,
            **kwargs
        )

        # initialize arrays in memory and fill with nans to start
        # this makes it easier to see bugs/wrong predictions than filling zeros
        preds = np.empty((len(loader.dataset), 3))
        preds[:] = np.nan

        all_labels = np.empty(len(loader.dataset))
        all_labels[:] = np.nan

        # save probs 
        probs = np.empty((len(loader.dataset), 3))
        probs[:] = np.nan

        prev_network_state = self.network.training

        # batch size might differ if user passes in a dataloader
        batch_size = loader.batch_size
        for idx, X in enumerate(tqdm(loader)):
            # Some dataloaders will have all_labels, handle this case
            top_probs, top_preds, label = self.predict_step(batch=X, batch_idx=idx)
            all_labels[idx * batch_size : (idx + 1) * batch_size] = label
            preds[idx * batch_size : (idx + 1) * batch_size] = top_preds
            probs[idx * batch_size : (idx + 1) * batch_size] = top_probs

        preds = pd.DataFrame(preds).astype(int)
        preds = preds.rename(
            {
                0: "first_pred",
                1: "second_pred",
                2: "third_pred",
            },
            axis=1,
        )

        preds = preds.apply(lambda x: self.label_encoder.inverse_transform(x))

        probs = pd.DataFrame(probs)
        probs = probs.rename(
            {
                0: "first_prob",
                1: "second_prob",
                2: "third_prob",
            },
            axis=1,
        )

        final = pd.concat([preds, probs], axis=1)

        if not np.all(np.isnan(all_labels)):
            final["label"] = all_labels


        # if network was in training mode before inference, set it back to that
        if prev_network_state:
            self.network.train()

        return final

    def predict_step(self, batch: Any, batch_idx: int, dataloader_idx: int = 0) -> Any:
        if len(batch) == 2:
            data, label = batch
        else:
            data, label = batch, None
        data = data.float()
        res = self(data)[0]
        probs, top_preds = res.topk(3, axis=1)  # to get indices
        probs = probs.softmax(dim=-1)

        return probs.detach().cpu().numpy(), top_preds.detach().cpu().numpy(), label
 
    def temperature_scale(self, logits):
        """
        Perform temperature scaling on logits
        """
        # Expand temperature to match the size of logits
        temperature = self.temperature.unsqueeze(1).expand(logits.size(0), logits.size(1))  # (Batch, Classes)
        return logits / temperature  # (Batch, Classes)

    def set_temperature(self, dataloader, max_iter=50, lr=0.01):
        """
        Tune the temperature of the model (using the validation set).
        We're going to set it to optimize NLL.
        dataloader (DataLoader): validation set loader
        """
        nll_criterion = torch.nn.CrossEntropyLoss()
        ece_criterion = _ECELoss()

        # First: collect all the logits and labels for the validation set
        logits_list = []
        labels_list = []
        print("Setting temperature ...")
        with torch.no_grad():
            for data, label in tqdm(dataloader):
                logits = self(data)[0]
                logits_list.append(logits)
                labels_list.append(label)
            logits = torch.cat(logits_list) # (num_samples*batch_size, num_classes)
            labels = torch.cat(labels_list)

        # Calculate NLL and ECE before temperature scaling
        before_temperature_nll = nll_criterion(logits, labels).item()
        before_temperature_ece = ece_criterion(logits, labels).item()
        print('Before temperature - NLL: %.3f, ECE: %.3f' % (before_temperature_nll, before_temperature_ece))

        # Next: optimize the temperature w.r.t. NLL
        optimizer = torch.optim.LBFGS([self.temperature], lr=lr, max_iter=max_iter)

        def eval():
            optimizer.zero_grad()
            loss = nll_criterion(self.temperature_scale(logits), labels)
            loss.backward()
            return loss

        optimizer.step(eval)
        
        after_temperature_nll = nll_criterion(self.temperature_scale(logits), labels).item()
        after_temperature_ece = ece_criterion(self.temperature_scale(logits), labels).item()
        print('Optimal temperature: %.3f' % self.temperature.item())
        print('After temperature - NLL: %.3f, ECE: %.3f' % (after_temperature_nll, after_temperature_ece))

        return self

def confusion_matrix(model, dataloader, num_classes):
    confusion_matrix = torch.zeros(num_classes, num_classes)
    with torch.no_grad():
        for i, (inputs, classes) in enumerate(tqdm(dataloader)):
            outputs, _ = model(inputs)

            _, preds = torch.max(outputs, 1)
            for t, p in zip(classes.view(-1), preds.view(-1)):
                confusion_matrix[t.long(), p.long()] += 1

    return confusion_matrix


def median_f1(tps, fps, fns):
    precisions = tps / (tps + fps)
    recalls = tps / (tps + fns)

    f1s = 2 * (np.dot(precisions, recalls)) / (precisions + recalls)

    return np.nanmedian(f1s)


def aggregate_metrics(num_classes) -> Dict[str, Callable]:
    task = "binary" if num_classes == 2 else "multiclass"
    num_classes = None if num_classes == 2 else num_classes

    metrics = {
        "micro_accuracy": Accuracy(task=task, num_classes=num_classes, average="micro"),
        "macro_accuracy": Accuracy(task=task, num_classes=num_classes, average="macro"),
        "weighted_accuracy": Accuracy(task=task, num_classes=num_classes, average="weighted"),
        "precision": Precision(task=task, num_classes=num_classes, average="macro"),
        "recall": Recall(task=task, num_classes=num_classes, average="macro"),
        "f1": F1Score(task=task, num_classes=num_classes, average="macro"),
        "specificity": Specificity(task=task, num_classes=num_classes, average="macro"),
    }

    return metrics






import linecache
from functools import cached_property
from typing import Any, List

import numpy as np
import pandas as pd
import torch
from scipy.sparse import issparse
from torch.utils.data import Dataset
import anndata as an
from typing import Union

class DatasetForInference(Dataset):
    def __init__(self, adata: Union[an.AnnData, np.ndarray]) -> None:
        super().__init__()
        self.data = adata.X if isinstance(adata, an.AnnData) else adata

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            it = list(range(idx.start or 0, idx.stop or len(self), idx.step or 1))
            return [self[i] for i in it]

        data = self.data[idx, :]
        if issparse(data):
            data = data.todense()

        data = torch.from_numpy(data)
        return data

    def __len__(self):
        return self.data.shape[0]

    @property
    def shape(self):
        return self.data.shape


from .data import *
from .inference import DatasetForInference
from .lightning_train import *
from .model import *
from .networking import UploadCallback
from .scvi_api import SIMS
from .temperature_scaling import *

__all__ = [
    "SIMS",
    "SIMSClassifier",
    "DataModule",
    "UploadCallback",
    "DatasetForInference"
]



import pathlib
from functools import cached_property
from typing import Optional

import anndata as an
import numpy as np
import pandas as pd
import pytorch_lightning as pl
import torch
from sklearn.preprocessing import LabelEncoder

from .data import generate_dataloaders
from sklearn.utils.class_weight import compute_class_weight

here = pathlib.Path(__file__).parent.absolute()

class DataModule(pl.LightningDataModule):
    def __init__(
        self,
        data: an.AnnData = None,
        class_label: str = None,
        test_size: float = 0.2,
        batch_size = 32,
        num_workers = 0,
        device=("cuda:0" if torch.cuda.is_available() else None),
        *args,
        **kwargs,
    ):
        super().__init__()
        self.device = device
        self.class_label = class_label
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.data = data
        self.test_size = test_size

        self.args = args
        self.kwargs = kwargs

        self.prepared = False
        self.setuped = False

        if data is not None:
            self.prepare_data()
            self.setup()

    def prepare_data(self):
        unique_targets = self.data.obs.loc[:, self.class_label].unique()
        label_encoder = LabelEncoder().fit(unique_targets)

        self.label_encoder = label_encoder

        if not pd.api.types.is_numeric_dtype(self.data.obs.loc[:, self.class_label]):
            print("Numerically encoding class labels")
            self.data.obs.loc[:, f"numeric_{self.class_label}"] = self.label_encoder.transform(
                self.data.obs.loc[:, self.class_label]
            ).astype(int)

            self.class_label = f"numeric_{self.class_label}"

    def setup(self, stage: Optional[str] = None):
        loaders = generate_dataloaders(
            data=self.data,
            class_label=self.class_label,
            test_size=self.test_size,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            split=self.test_size is not None,
            *self.args,
            **self.kwargs,
        )

        if len(loaders) == 1:
            self.trainloader = loaders[0]
            self.valloader = None
            self.testloader = None
        else: 
            self.trainloader, self.valloader, self.testloader = loaders

        print("Calculating weights")
        labels = self.data.obs.loc[:, self.class_label].values
        self.weights = torch.from_numpy(
            compute_class_weight(
                y=labels,
                classes=np.unique(labels),
                class_weight="balanced",
            )
        ).float().to(self.device)


        self.setuped = True

    def train_dataloader(self):
        return self.trainloader

    def val_dataloader(self):
        return self.valloader

    def test_dataloader(self):
        return self.testloader

    @cached_property
    def num_labels(self):
        return self.data.obs.loc[:, self.class_label].nunique()

    @cached_property
    def num_features(self):
        return self.data.shape[1]

    @cached_property
    def input_dim(self):
        return self.num_features

    @cached_property
    def output_dim(self):
        return self.num_labels

    @cached_property
    def genes(self):
        return self.data.var.index.tolist()
    
    @cached_property
    def cells(self):
        return self.data.obs.index.tolist()


import os
import sys

import numpy as np
import pandas as pd
import torch
from pytorch_tabnet.tab_model import TabNetClassifier
from torch.utils.data import *
from tqdm import tqdm

sys.path.append("../src/")
sys.path.append("..")

from src.helper import *
from src.models.lib.data import *


def map_cols_test():
    ref = ["a", "b", "c"]
    curr = ["b", "a", "c", "d"]
    sample = np.array([1, 2, 3, 4])  # Want --> [2,1,3]

    result = clean_sample(sample, ref, curr)
    desired = torch.from_numpy(np.array([2, 1, 3]))

    assert torch.equal(result, desired)

    ref = ["a", "b", "c"]
    curr = ["c", "d", "b", "a"]

    sample = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
    # --> want [[4, 3, 1],
    #           [8, 7, 5]]

    res = clean_sample(sample, ref, curr)
    desired = torch.from_numpy(np.array([[4, 3, 1], [8, 7, 5]]))

    assert torch.equal(res, desired)


def _test_first_n_samples(n, datafile, labelfile, class_label="Type", index_col="cell"):
    data = GeneExpressionData(
        datafile,
        labelfile,
        class_label,
        skip=3,
        index_col=index_col,
    )
    cols = data.columns

    # Generate dict with half precision values to read this into my 16gb memory
    dtype_cols = dict(zip(cols, [np.float32] * len(cols)))

    data_df = pd.read_csv(
        datafile, nrows=2 * n, header=1, dtype=dtype_cols
    )  # Might need some extras since numerical index drops some values
    label_df = pd.read_csv(labelfile, nrows=n)

    similar = []
    for i in range(n):
        datasample = data[i][0]

        dfsample = torch.from_numpy(data_df.loc[label_df.loc[i, index_col], :].values).float()
        isclose = all(torch.isclose(datasample, dfsample))
        similar.append(isclose)

    print(f"First {n=} columns of expression matrix is equal to GeneExpressionData: {all(p for p in similar)}")

    assert all(p for p in similar)


def test_split(n, datafile, labelfile, class_label="Type", index_col="cell"):
    from sklearn.model_selection import train_test_split

    # 2*n is usually enough to capture differences in index_col

    label_df = pd.read_csv(labelfile, nrows=2 * n)
    data_df = pd.read_csv(datafile, nrows=2 * n, dtype=np.float32, header=1)

    labels = label_df.loc[:, class_label]

    train, _ = train_test_split(labels, random_state=42)
    train = train.index

    train_labeldf = label_df.loc[train, :].reset_index(drop=True)

    train = GeneExpressionData(
        datafile,
        labelfile,
        index_col=index_col,
        class_label=class_label,
        indices=train,
    )

    train_similar = []
    for i in range(n):
        trainsample = train[i][0]
        df_idx = train_labeldf.loc[i, index_col]

        trainsample_fromdf = torch.from_numpy(data_df.loc[df_idx, :].values)

        train_similar.append(all(torch.isclose(trainsample, trainsample_fromdf)))

    assert all(p for p in train_similar)  # all train are similar


def test_datamodule():
    N = 50
    datafiles, labelfiles = list(INTERIM_DATA_AND_LABEL_FILES_LIST.keys()), list(INTERIM_DATA_AND_LABEL_FILES_LIST.values())
    datafiles = [os.path.join("..", "data", "interim", f) for f in datafiles]
    processed_labels = [os.path.join("..", "data", "processed/labels", f) for f in labelfiles]

    for datafile, labelfile in zip(datafiles, processed_labels):
        print(f"Testing {datafile=}")
        _test_first_n_samples(N, datafile, labelfile)


def test_retina_data():
    N = 50
    datafile = os.path.join("..", "data", "retina", "retina_T.csv")
    labelfile = os.path.join("..", "data", "retina", "retina_labels_numeric.csv")

    _test_first_n_samples(N, datafile, labelfile, "class_label", "cell")
    test_split(N, datafile, labelfile, "class_label", "cell")


if __name__ == "__main__":
    map_cols_test()
    test_datamodule()
    test_retina_data()



import os
import pathlib
import random
import sys
import unittest

import anndata as an
import numpy as np
import pandas as pd
import torch
from torch.utils.data import *

random.seed(42)
from os.path import abspath, dirname, join

sys.path.append(join(dirname(abspath(__file__)), "..", "src"))

from data import *
from network import *


def generate_synthetic_h5ad():
    df = pd.DataFrame(index=range(25), columns=[f"col_{i}" for i in range(10)])

    # Such that row_i = (i, ... ,i)
    for i in range(25):
        df.loc[i, :] = [i] * 10

    # Convert this to an anndata object
    cols = df.columns

    df = an.AnnData(df.values)
    df.var.index = cols

    # Define label dataframe
    labels = pd.DataFrame(index=range(25), columns=["index_col", "label"])

    # Create fake index_col = label col such that index_col_i = label_col_i = row_i = (i,...,i)
    labels = pd.DataFrame(index=range(25), columns=["index_col", "label"])
    labels["index_col"] = [int(x) for x in random.sample(list(df.obs.index), k=len(df))]
    labels["label"] = labels["index_col"]

    return df, labels


def generate_synethic_csv():
    # Create a test dataframe
    df = pd.DataFrame(index=range(25), columns=[f"col_{i}" for i in range(10)])

    # Such that row_i = (i, ... ,i)
    for i in range(25):
        df.loc[i, :] = [i] * 10

    # Define label dataframe
    labels = pd.DataFrame(index=range(25), columns=["index_col", "label"])

    # Create fake index_col = label col such that index_col_i = label_col_i = row_i = (i,...,i)
    labels["index_col"] = [int(x) for x in random.sample(list(df.index), k=len(df))]
    labels["label"] = labels["index_col"]

    return df, labels


class TestData(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.here = pathlib.Path(__file__).parent.resolve()

        cls.datapath = os.path.join(cls.here, "datapath")
        if not os.path.isdir(cls.datapath):
            print(f"Making test directories")
            os.makedirs(cls.datapath, exist_ok=True)

        # Define test file locations
        cls.datafile_csv = os.path.join(cls.datapath, "test_data.csv")
        cls.datafile_h5ad = os.path.join(cls.datapath, "test_data.h5ad")
        cls.labelfile = os.path.join(cls.datapath, "test_labels.csv")

    def test_dataset_from_csv(self):
        df, labels = generate_synethic_csv()

        labels.to_csv(self.labelfile, index=False)
        df.to_csv(self.datafile_csv)

        # Now, write the actual tests
        train, val, test = generate_single_dataset(
            datafile=self.datafile_csv,
            labelfile=self.labelfile,
            class_label="label",
            index_col="index_col",
            skip=2,
            stratify=False,
        )

        # Test train
        for i in range(len(train)):
            X, y = train[i]
            self.assertEqual(X[0], y)

        # Test val
        for i in range(len(val)):
            X, y = val[i]
            self.assertEqual(X[0], y)

        # Test test
        for i in range(len(test)):
            X, y = test[i]
            self.assertEqual(X[0], y)

    def test_dataset_from_h5ad(self):
        # Create a test dataframe
        df, labels = generate_synthetic_h5ad()

        labels.to_csv(self.labelfile, index=False)
        df.write_h5ad(self.datafile_h5ad)

        # Now, write the actual tests
        # Don't stratify since we have len(rows) = card(labels)
        train, val, test = generate_single_dataset(
            datafile=self.datafile_h5ad,
            labelfile=self.labelfile,
            class_label="label",
            index_col="index_col",
            stratify=False,
        )

        # Test train
        for i in range(len(train)):
            X, y = train[i]
            self.assertEqual(X[0], y)

        # Test val
        for i in range(len(val)):
            X, y = val[i]
            self.assertEqual(X[0], y)

        # Test test
        for i in range(len(test)):
            X, y = test[i]
            self.assertEqual(X[0], y)

    def test_refgene_alignment(self):
        ref = ["a", "b", "c"]
        curr = ["b", "a", "c", "d"]
        sample = np.array([1, 2, 3, 4])  # Want --> [2,1,3]

        result = clean_sample(sample, ref, curr)
        desired = torch.from_numpy(np.array([2, 1, 3]))

        assert torch.equal(result, desired)

        ref = ["a", "b", "c"]
        curr = ["c", "d", "b", "a"]

        sample = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
        # --> want [[4, 3, 1],
        #           [8, 7, 5]]

        res = clean_sample(sample, ref, curr)
        desired = torch.from_numpy(np.array([[4, 3, 1], [8, 7, 5]]))

        assert torch.equal(res, desired)

    def test_dataloader_from_csv(self):
        train, val, test = generate_dataloaders(
            datafiles=[self.datafile_csv],
            labelfiles=[self.labelfile],
            class_label="label",
            index_col="index_col",
            skip=2,
            stratify=False,
        )

        for sample in train:
            X, Y = sample[0], sample[1]
            for x, y in zip(X, Y):
                self.assertEqual(x[0], y)

        for sample in val:
            X, Y = sample[0], sample[1]
            for x, y in zip(X, Y):
                self.assertEqual(x[0], y)

        for sample in test:
            X, Y = sample[0], sample[1]
            for x, y in zip(X, Y):
                self.assertEqual(x[0], y)

    def test_dataloader_from_h5ad(self):
        train, val, test = generate_dataloaders(
            datafiles=[self.datafile_h5ad],
            labelfiles=[self.labelfile],
            class_label="label",
            index_col="index_col",
            stratify=False,
        )

        for sample in train:
            X, Y = sample[0], sample[1]
            for x, y in zip(X, Y):
                self.assertEqual(x[0], y)

        for sample in val:
            X, Y = sample[0], sample[1]
            for x, y in zip(X, Y):
                self.assertEqual(x[0], y)

        for sample in test:
            X, Y = sample[0], sample[1]
            for x, y in zip(X, Y):
                self.assertEqual(x[0], y)


if __name__ == "__main__":
    unittest.main()



import pathlib
from typing import Union

import anndata as an
import pytorch_lightning as pl
from pytorch_lightning.callbacks import ModelCheckpoint, Timer, EarlyStopping

from scsims.lightning_train import DataModule
from scsims.model import SIMSClassifier
import pandas as pd 
from sklearn.preprocessing import LabelEncoder


here = pathlib.Path(__file__).parent.absolute()

class UnconfiguredModelError(Exception):
    pass

class SIMS:
    def __init__(
        self,
        *,
        weights_path: str = None,
        data: an.AnnData = None,
        class_label: str = None,
        **kwargs,
    ) -> None:
        if weights_path is not None:
            self.model = SIMSClassifier.load_from_checkpoint(weights_path, **kwargs, strict=False)

        self.datamodule = DataModule(data=data, class_label=class_label, **kwargs)

        for att, value in self.datamodule.__dict__.items():
            setattr(self, att, value)

    def setup_model(self, *args, **kwargs):
        print('Setting up model ...')
        if 'model_size' in kwargs:
            assert kwargs['model_size'] in ["tall", "grande", "venti", "trenta"], "if specified, model_size must be one of 'big', 'medium', or 'small'"
            if kwargs['model_size'] == "trenta":
                kwargs['n_a'] = 128
                kwargs['n_d'] = 128
            if kwargs['model_size'] == "venti":
                kwargs['n_a'] = 64
                kwargs['n_d'] = 64
            if kwargs['model_size'] == "grande":
                kwargs['n_a'] = 32
                kwargs['n_d'] = 32
            if kwargs['model_size'] == "tall":
                kwargs['n_a'] = 8
                kwargs['n_d'] = 8

        self.model = SIMSClassifier(
            input_dim=self.datamodule.input_dim, 
            output_dim=self.datamodule.output_dim, 
            genes=self.datamodule.genes,
            cells=self.datamodule.cells,
            label_encoder=self.datamodule.label_encoder,
            *args, 
            **kwargs
        )

    def setup_trainer(self, early_stopping_patience: int = None, *args, **kwargs):
        print('Setting up trainer ...')
        if 'callbacks' in kwargs:
            # check if any of the list of callbacks is a modelcheckpoint or timer
            callbacks = kwargs['callbacks']
            if not any([isinstance(callback, ModelCheckpoint) for callback in callbacks]):
                callbacks.append(ModelCheckpoint())
            if not any([isinstance(callback, Timer) for callback in callbacks]):
                callbacks.append(Timer())
            if not any([isinstance(callback, EarlyStopping) for callback in callbacks]) and early_stopping_patience is not None:
                callbacks.append(EarlyStopping(monitor='val_loss', patience=early_stopping_patience))
        else:
            kwargs['callbacks'] = [ModelCheckpoint(), Timer()]
            if early_stopping_patience is not None:
                kwargs['callbacks'].append(EarlyStopping(monitor='val_loss', patience=early_stopping_patience))
        self._trainer = pl.Trainer(*args, **kwargs)

    def train(self, *args, **kwargs):
        print('Beginning training')
        if not hasattr(self, "_trainer"):
            self.setup_trainer(*args, **kwargs)
        if not hasattr(self, "model"):
            self.setup_model(*args, **kwargs)

        self._trainer.fit(self.model, datamodule=self.datamodule)

        print('Finished training')

    def predict(self, inference_data: an.AnnData, *args, **kwargs):
        if not hasattr(self, 'model'):
            raise UnconfiguredModelError(
                """The model attribute is not configured. This is likely 
                because you are running the predict method after re-initializing the 
                SIMS class. Reinitialize the SIMS class with the weights_path
                pointing to the .ckpt file to continue."""
            )
        self.results = self.model.predict(inference_data, *args, **kwargs)
        return self.results

    def explain(self, datafiles: an.AnnData, labelfile=None, class_label=None, *args, **kwargs):
        print('Computing explainability matrix ...')
        results = self.model.explain(datafiles, *args, **kwargs)

        return results

    def decode_predictions(self, labelfile: str, class_label: str):
        labels = pd.read_csv(labelfile)
        label_encoder = LabelEncoder()
        label_encoder.fit(labels[class_label])
        results = self.results
        try:
            pred_columns = [col for col in results.columns if 'pred' in col]
            results[pred_columns] = results[pred_columns].apply(lambda x: label_encoder.inverse_transform(x))
            #results = self.results.apply(lambda x: label_encoder.inverse_transform(x))
        except AttributeError:
            raise AttributeError(
                """The results attribute is not configured. This is likely 
                because you are running the decode_predictions method before running the predict method.
                Run the predict method first, then run the decode_predictions method."""
            )
        return results

    def set_temperature(self, dataloader):
        self.model = self.model.set_temperature(dataloader)


import inspect
import linecache
import pathlib
import warnings
from functools import cache, cached_property, partial
from itertools import chain
from typing import *

import anndata as an
import numpy as np
import torch
from scipy.sparse import issparse
from sklearn.model_selection import train_test_split
from torch.utils.data import DataLoader, Dataset

class TransformSequence:
    def __init__(self, *transforms):
        self.transforms = transforms

    def __call__(self, x):
        for transform in self.transforms:
            x = transform(x)
        return x

class RandomNoise:
    def __init__(self, mean=0, std=0.1):
        self.mean = mean
        self.std = std

    def __call__(self, x):
        return x + torch.randn_like(x) * self.std + self.mean

class RandomDropout:
    def __init__(self, p=0.2):
        self.p = p

    def __call__(self, x):
        return torch.nn.functional.dropout(x, p=self.p, training=True)

class AnnDatasetMatrix(Dataset):
    def __init__(
        self,
        matrix: np.ndarray,
        labels: List[any],
        split: Collection[int] = None,
        transform: TransformSequence = None, 
        *args,
        **kwargs,
    ) -> None:
        super().__init__()
        self.data = matrix
        self.labels = labels
        self.split = split
        self.transform = transform

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            step = 1 if idx.step is None else idx.step
            idxs = range(idx.start, idx.stop, step)
            return [self[i] for i in idxs]

        data = self.data[idx, :]

        if issparse(data):
            data = data.todense()
            # Need to get first row of 1xp matrix, weirdly this is how to do it :shrug:
            data = np.squeeze(np.asarray(data))

        if self.transform is not None:
            data = self.transform(data)

        return (torch.from_numpy(data), self.labels[idx])

    def __len__(self):
        return self.data.shape[0] if issparse(self.data) else len(self.data)

    @property
    def shape(self):
        return self.data.shape


class CollateLoader(DataLoader):
    def __init__(
        self,
        dataset: Type[Dataset],
        batch_size: int,
        num_workers: int,
        transpose: bool = False,
        normalize: bool = True,
        *args,
        **kwargs,
    ) -> None:
        collate_fn = partial(
            _standard_collate,
            normalize=normalize,
            transpose=transpose,
        )

        # This is awkward, but Dataloader init doesn't handle optional keyword arguments
        # So we have to take the intersection between the passed **kwargs and the DataLoader named arguments
        allowed_args = inspect.signature(super().__init__).parameters
        new_kwargs = {}

        for key in allowed_args:
            name = allowed_args[key].name
            if name in kwargs:
                new_kwargs[key] = kwargs[key]

        # Finally, initialize the DataLoader
        super().__init__(
            dataset=dataset,
            collate_fn=collate_fn,
            batch_size=batch_size,
            num_workers=num_workers,
            **new_kwargs,
        )


class SequentialLoader:
    """
    Class to sequentially stream samples from an arbitrary number of DataLoaders.

    :param dataloaders: List of DataLoaders or DataLoader derived class, such as the CollateLoader from above
    :type dataloaders: List[Union[DataLoader, SequentialLoader]]
    """

    def __init__(self, dataloaders):
        self.dataloaders = dataloaders

    def __len__(self):
        return sum(len(dl) for dl in self.dataloaders)

    def __iter__(self):
        yield from chain(*self.dataloaders)

def _standard_collate(
    sample: List[tuple],
    normalize: bool,
    transpose: bool,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Collate minibatch of samples, optionally normalizing and transposing.

    :param sample: List of DelimitedDataset items to collate
    :type sample: List[tuple]
    :param normalize: boolean, indicates if we should transpose the minibatch (in the case of incorrectly formatted .csv data)
    :type normalize: bool
    :param transpose: boolean, indicates if we should normalize the minibatch
    :type transpose: bool
    :return: Collated samples and labels, respectively
    :rtype: Tuple[torch.Tensor, torch.Tensor]
    """

    if len(sample[0]) == 2:
        data = torch.stack([x[0] for x in sample])
        labels = torch.tensor([x[1] for x in sample])

        return _transform_sample(data, normalize, transpose), labels
    else:  # len == 1
        return _transform_sample(torch.stack(sample), normalize, transpose)


def _transform_sample(data: torch.Tensor, normalize: bool, transpose: bool) -> torch.Tensor:
    if transpose:
        data = data.T
    if normalize:
        data = torch.nn.functional.normalize(data)

    return data

def clean_sample(
    sample: torch.Tensor,
    refgenes: List[str],
    currgenes: List[str],
) -> torch.Tensor:

    indices = np.intersect1d(currgenes, refgenes, return_indices=True)[1]

    if sample.ndim == 2:
        sample = sample[:, indices]
    else:
        sample = sample[indices]  # in the case of a 1d array (single row)

    return sample


def generate_dataloaders(
    data: an.AnnData,
    class_label: str,
    test_prop=0.2,
    stratify=True,
    batch_size: int = 16,
    num_workers: int = 0,
    split: bool = True,
    deterministic: bool = True,
    *args,
    **kwargs,
) -> Union[Tuple[CollateLoader, List[CollateLoader]], Tuple[SequentialLoader]]:
    if isinstance(data, str):
        data = an.read_h5ad(data)

    current_labels = data.obs.loc[:, class_label]

    # make sure data can be stratified
    if stratify:
        if len(current_labels.unique()) < 3:
            warnings.warn(
                "One class has less than 3 samples, disabling stratification"
            )
            stratify = False

    datasets = []
    if split:
        trainsplit, valsplit = train_test_split(
            current_labels,
            stratify=(current_labels if stratify else None),
            test_size=test_prop,
            random_state=(42 if deterministic else None),
        )

        trainsplit, testsplit = train_test_split(
            trainsplit,
            stratify=(trainsplit if stratify else None),
            test_size=test_prop,
            random_state=(42 if deterministic else None),
        )

        for split in [trainsplit, valsplit, testsplit]:
            datasets.append(
                AnnDatasetMatrix(
                    matrix=data[split.index, :].X,
                    labels=split.values,
                    split=split.index,
                )
            )
    else:
        train = AnnDatasetMatrix(
            matrix=data.X,
            labels=current_labels.values,
            split=current_labels.index,
            *args,
            **kwargs,
        )
        datasets.append(train)

    loaders = [
        CollateLoader(
            dataset=dataset,
            batch_size=batch_size,
            num_workers=num_workers,
            *args,
            **kwargs,
        )
        for dataset in datasets
    ]

    return loaders




import os
import pathlib
import random
import sys
import unittest

import anndata as an
import numpy as np
import pandas as pd
import pytorch_lightning as pl
import sklearn as sk
import torch
from torch.utils.data import *

random.seed(42)
from os.path import abspath, dirname, join

sys.path.append(join(dirname(abspath(__file__)), "..", "src"))  # hacky, to fix pathing

from data import *
from lightning_train import *
from model import *


class TestModel(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.here = pathlib.Path(__file__).parent.resolve()

        cls.datapath = os.path.join(cls.here, "datapath")
        if not os.path.isdir(cls.datapath):
            print(f"Making test directories")
            os.makedirs(cls.datapath, exist_ok=True)

        # Define test file locations
        cls.datafile = os.path.join(cls.datapath, "test_data_pipeline.csv")
        cls.labelfile = os.path.join(cls.datapath, "test_data_pipeline.csv")

    def train_pipeline(self):
        # Generate synthetic data
        data = sk.datasets.make_classification()

        df, labels = data
        df = pd.DataFrame(df)
        labels = pd.DataFrame(df)
        labels.index.name = "class_label"

        df.to_csv(self.datafile, index=False)
        labels.to_csv(self.labelfile, index=False)

        module = DataModule(
            datafiles=self.datafile,
            labelfiles=self.labelfile,
            class_label="class_label",
            batch_size=4,
            num_workers=0,
        )

        model = SIMSClassifier(
            input_dim=module.num_features,
            output_dim=module.num_labels,
        )

        trainer = pl.Trainer(
            gpus=(1 if torch.cuda.is_available() else 0),
            auto_lr_find=False,
            max_epochs=1,
        )

        trainer.fit(model, datamodule=module)


if __name__ == "__main__":
    unittest.main()



from typing import *

import numpy as np
import pytorch_lightning as pl
import torch
from pytorch_tabnet.tab_network import (EmbeddingGenerator, RandomObfuscator,
                                        TabNetDecoder, TabNetEncoder)


# ALL THIS IS FLATTENING THE API FROM https://github.com/dreamquark-ai/tabnet
class NoiseObfuscator(torch.nn.Module):
    def __init__(
        self,
        variance: float = 1,
    ) -> None:
        """
        Custom dataset interiting from parent Dataset that adds Gaussian noise with a given variance to each sample, to be used for pretraining

        :param variance: Variance of Gaussian noise, defaults to 1
        :type variance: float, optional
        """

        self.variance = variance

    def forward(self, x):
        return x + self.variance * torch.randn_like(x)


def UnsupervisedLoss(y_pred, embedded_x, obf_vars, eps=1e-9):
    """
    Implements unsupervised loss function.
    This differs from orginal paper as it's scaled to be batch size independent
    and number of features reconstructed independent (by taking the mean)
    Parameters
    ----------
    y_pred : torch.Tensor or np.array
        Reconstructed prediction (with embeddings)
    embedded_x : torch.Tensor
        Original input embedded by network
    obf_vars : torch.Tensor
        Binary mask for obfuscated variables.
        1 means the variable was obfuscated so reconstruction is based on this.
    eps : float
        A small floating point to avoid ZeroDivisionError
        This can happen in degenerated case when a feature has only one value
    Returns
    -------
    loss : torch float
        Unsupervised loss, average value over batch samples.
    """
    errors = y_pred - embedded_x
    reconstruction_errors = torch.mul(errors, obf_vars) ** 2
    batch_means = torch.mean(embedded_x, dim=0)
    batch_means[batch_means == 0] = 1

    batch_stds = torch.std(embedded_x, dim=0) ** 2
    batch_stds[batch_stds == 0] = batch_means[batch_stds == 0]
    features_loss = torch.matmul(reconstruction_errors, 1 / batch_stds)
    # compute the number of obfuscated variables to reconstruct
    nb_reconstructed_variables = torch.sum(obf_vars, dim=1)
    # take the mean of the reconstructed variable errors
    features_loss = features_loss / (nb_reconstructed_variables + eps)
    # here we take the mean per batch, contrary to the paper
    loss = torch.mean(features_loss)
    return loss


def UnsupervisedLossNumpy(y_pred, embedded_x, obf_vars, eps=1e-9):
    """Compute Euclidean distance between reconstructed and original vector

    :param y_pred: _description_
    :type y_pred: _type_
    :param embedded_x: _description_
    :type embedded_x: _type_
    :param obf_vars: _description_
    :type obf_vars: _type_
    :param eps: _description_, defaults to 1e-9
    :type eps: _type_, optional
    :return: _description_
    :rtype: _type_
    """
    errors = y_pred - embedded_x
    reconstruction_errors = np.multiply(errors, obf_vars) ** 2
    batch_means = np.mean(embedded_x, axis=0)
    batch_means = np.where(batch_means == 0, 1, batch_means)

    batch_stds = np.std(embedded_x, axis=0, ddof=1) ** 2
    batch_stds = np.where(batch_stds == 0, batch_means, batch_stds)
    features_loss = np.matmul(reconstruction_errors, 1 / batch_stds)
    # compute the number of obfuscated variables to reconstruct
    nb_reconstructed_variables = np.sum(obf_vars, axis=1)
    # take the mean of the reconstructed variable errors
    features_loss = features_loss / (nb_reconstructed_variables + eps)
    # here we take the mean per batch, contrary to the paper
    loss = np.mean(features_loss)
    return loss


class TabNetPretraining(pl.LightningModule):
    def __init__(
        self,
        input_dim,
        pretraining_ratio=0.2,
        n_d=8,
        n_a=8,
        n_steps=3,
        gamma=1.3,
        cat_idxs=[],
        cat_dims=[],
        cat_emb_dim=1,
        n_independent=2,
        n_shared=2,
        epsilon=1e-15,
        virtual_batch_size=128,
        momentum=0.02,
        mask_type="sparsemax",
        n_shared_decoder=1,
        n_indep_decoder=1,
    ):
        super(TabNetPretraining, self).__init__()

        self.cat_idxs = cat_idxs or []
        self.cat_dims = cat_dims or []
        self.cat_emb_dim = cat_emb_dim

        self.input_dim = input_dim
        self.n_d = n_d
        self.n_a = n_a
        self.n_steps = n_steps
        self.gamma = gamma
        self.epsilon = epsilon
        self.n_independent = n_independent
        self.n_shared = n_shared
        self.mask_type = mask_type
        self.pretraining_ratio = pretraining_ratio
        self.n_shared_decoder = n_shared_decoder
        self.n_indep_decoder = n_indep_decoder

        if self.n_steps <= 0:
            raise ValueError("n_steps should be a positive integer.")
        if self.n_independent == 0 and self.n_shared == 0:
            raise ValueError("n_shared and n_independent can't be both zero.")

        self.virtual_batch_size = virtual_batch_size
        self.embedder = EmbeddingGenerator(input_dim, cat_dims, cat_idxs, cat_emb_dim)
        self.post_embed_dim = self.embedder.post_embed_dim

        self.masker = RandomObfuscator(self.pretraining_ratio)
        self.encoder = TabNetEncoder(
            input_dim=self.post_embed_dim,
            output_dim=self.post_embed_dim,
            n_d=n_d,
            n_a=n_a,
            n_steps=n_steps,
            gamma=gamma,
            n_independent=n_independent,
            n_shared=n_shared,
            epsilon=epsilon,
            virtual_batch_size=virtual_batch_size,
            momentum=momentum,
            mask_type=mask_type,
        )

        self.decoder = TabNetDecoder(
            self.post_embed_dim,
            n_d=n_d,
            n_steps=n_steps,
            n_independent=self.n_indep_decoder,
            n_shared=self.n_shared_decoder,
            virtual_batch_size=virtual_batch_size,
            momentum=momentum,
        )

    def forward(self, x):
        embedded_x = self.embedder(x)
        if self.training:
            masked_x, obf_vars = self.masker(embedded_x)
            # set prior of encoder with obf_mask
            prior = 1 - obf_vars
            steps_out, _ = self.encoder(masked_x, prior=prior)
            res = self.decoder(steps_out)
            return res, embedded_x, obf_vars
        else:
            steps_out, _ = self.encoder(embedded_x)
            res = self.decoder(steps_out)
            return res, embedded_x, torch.ones(embedded_x.shape).to(x.device)

    def forward_masks(self, x):
        embedded_x = self.embedder(x)
        return self.encoder.forward_masks(embedded_x)

    def training_step(self, batch, batch_idx):
        y, y_hat, loss = self._step(batch)

        self.log("train_loss", loss, logger=True, on_epoch=True, on_step=True)
        self._compute_metrics(y_hat, y, "train")

        return loss

    def _compute_metrics(
        self,
        y_hat: torch.Tensor,
        y: torch.Tensor,
        tag: str,
        on_epoch=True,
        on_step=False,
    ):
        val = metric(y_hat, y, average="weighted", num_classes=self.output_dim)

        self.log(
            f"Reconstuction loss",
            val,
            on_epoch=on_epoch,
            on_step=on_step,
            logger=True,
        )


def pretrain_model(
    datamodule,
    model,
    trainer,
):
    """
    Pretrain model via random masking or denoising

    :param datamodule: Datamodule to train model on
    :type datamodule: pl.DataModule
    :param model: Model to train on
    :type model: Model
    :param trainer: PyTorch Lightning trainer used to train model
    :type trainer: pl.LightningTrainer
    """

    pass



#!/usr/bin/env python

"""The setup script."""

import pathlib

from setuptools import find_packages, setup

def read_requirements(file_path):
    with open(file_path, "r") as f:
        requirements = f.read().splitlines()
    return requirements

requirements = read_requirements("requirements.txt")
setup_requirements = requirements.copy()
test_requirements = []

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    author="Julian Lehrer",
    author_email="jmlehrer@ucsc.edu",
    python_requires=">=3.9",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Scalable, Interpretable Deep Learning for Single-Cell RNA-seq Classification",
    install_requires=requirements.copy(),
    license="MIT license",
    long_description=README,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="scsims",
    name="scsims",
    packages=find_packages(exclude=["tests"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/braingeneers/sims",
    version="3.0.1",
    zip_safe=False,
)



import numpy as np
import pytorch_lightning as pl
import torch
import torch.nn as nn


class Encoder(nn.Module):
    def __init__(
        self,
        input_dim=None,
        output_dim=None,
        layers=None,
    ) -> None:
        super().__init__()

        if all((input_dim, output_dim, layers)) == None:
            raise ValueError("If layers aren't specified, input and output dimensions must be")

        self.input_dim = input_dim
        self.output_dim = output_dim
        self.encode = (
            layers
            if layers is not None
            else nn.Sequential(
                nn.Linear(input_dim, 10000),
                nn.ReLU(),
                nn.Linear(10000, 5000),
                nn.ReLU(),
                nn.Linear(5000, 1000),
                nn.ReLU(),
                nn.Linear(1000, 500),
                nn.ReLU(),
                nn.Linear(500, output_dim),
            )
        )

    def forward(self, x):
        return self.encode(x)


class Decoder(nn.Module):
    def __init__(
        self,
        output_dim=None,
        input_dim=None,
        layers=None,
    ) -> None:
        super().__init__()

        if all((output_dim, input_dim, layers)) == None:
            raise ValueError("If layers aren't specified, output_dim and input_dim must be.")

        self.input_dim = input_dim
        self.output_dim = output_dim
        self.decode = (
            layers
            if layers is not None
            else nn.Sequential(
                nn.Linear(input_dim, 500),
                nn.ReLU(),
                nn.Linear(500, 1000),
                nn.ReLU(),
                nn.Linear(1000, 5000),
                nn.ReLU(),
                nn.Linear(5000, 10000),
                nn.ReLU(),
                nn.Linear(10000, output_dim),
            )
        )

    def forward(self, x):
        return self.decode(x)


class AutoEncoder(pl.LightningModule):
    def __init__(
        self,
        data_shape=None,
        encoder_layers=None,
        decoder_layers=None,
        optim_params=None,
        scheduler_params=None,
        loss=None,
    ) -> None:
        super().__init__()

        self.encoder = Encoder(input_dim=data_shape, layers=encoder_layers)
        self.decoder = Decoder(output_dim=data_shape, layers=decoder_layers)

        self.loss = loss if loss is not None else nn.MSELoss()
        self.optim_params = (
            optim_params
            if optim_params is not None
            else {
                "optimizer": torch.optim.Adam,
                "lr": 0.001,
                "weight_decay": 0.001,
            }
        )

        self.scheduler_params = (
            scheduler_params
            if scheduler_params is not None
            else {
                "scheduler": torch.optim.lr_scheduler.ReduceLROnPlateau,
                "factor": 0.75,  # Reduce LR by 25% on plateau
            }
        )

    def encode(self, x):
        return self.encoder(x)

    def decode(self, x):
        return self.decoder(x)

    def forward(self, x):
        x = self.encoder(x)
        x = self.decoder(x)

        return x

    def configure_optimizers(self):
        if "optimizer" in self.optim_params:
            optimizer = self.optim_params.pop("optimizer")
            optimizer = optimizer(self.parameters(), **self.optim_params)
        else:
            optimizer = torch.optim.Adam(self.parameters(), lr=0.01, weight_decay=1e-5)

        if self.scheduler_params is not None:
            scheduler = self.scheduler_params.pop("scheduler")
            scheduler = scheduler(optimizer, **self.scheduler_params)

        if self.scheduler_params is None:
            return optimizer

        return {
            "optimizer": optimizer,
            "lr_scheduler": scheduler,
            "monitor": "val_loss",
        }

    def _step(self, tag, batch):
        x_hat = self.forward(batch)
        loss = self.loss(batch, x_hat)

        self.log(tag, loss)

        return loss

    def training_step(self, batch, batch_idx):
        return self._step("train_loss", batch)

    def validation_step(self, batch, batch_idx):
        return self._step("val_loss", batch)

    def test_step(self, batch, batch_idx):
        return self._step("test_loss", batch)



import os
from typing import *

import boto3
import pytorch_lightning as pl


class UploadCallback(pl.callbacks.Callback):
    """Custom PyTorch callback for uploading model checkpoints to a S3 bucket.

    Parameters:
    path: Local path to folder where model checkpoints are saved
    desc: Description of checkpoint that is appended to checkpoint file name on save
    upload_prefix: Path in bucket/ to upload model checkpoints to, defaults to model_checkpoints
    """

    def __init__(
        self,
        path: str,
        desc: str,
        s3: boto3.resource,
        bucket: str,
        upload_prefix="model_checkpoints",
        n_epochs: int = 10,
        quiet: bool = False,
    ) -> None:
        super().__init__()
        self.path = path
        self.desc = desc

        self.s3 = s3
        self.bucket = bucket
        self.upload_prefix = upload_prefix
        self.epochs = n_epochs
        self.quiet = quiet

    def on_train_epoch_end(self, trainer, pl_module):
        epoch = trainer.current_epoch

        if epoch % self.epochs == 0:  # Save every ten epochs
            checkpoint = f"checkpoint-{epoch}-desc-{self.desc}.ckpt"
            checkpoint_path = os.path.normpath(os.path.join(self.path, checkpoint))

            if not self.quiet:
                print(f"Saving checkpoint on epoch {epoch} to {checkpoint_path}")

            trainer.save_checkpoint(checkpoint_path)

            if not self.quiet:
                print(f"Uploading checkpoint at epoch {epoch}")

            try:
                self.s3.Bucket(self.bucket).upload_file(
                    Filename=checkpoint_path,
                    Key=os.path.join(self.upload_prefix, checkpoint_path),
                )

            except Exception as e:
                print(f"Error when uploading on epoch {epoch}")
                print(e)



from scanpy.datasets import blobs 
from scsims import SIMS 
import numpy as np
import os 
import pytorch_lightning as pl 
import shutil

def test_inference_more_genes_than_train():
    traindata = blobs(n_variables=10, n_observations=100)
    testdata = blobs(n_variables=20, n_observations=100)
    
    # make data 32-bit
    traindata.X = traindata.X.astype(np.float32)
    testdata.X = testdata.X.astype(np.float32)

    model = SIMS(data=traindata, class_label="blobs")
    model.setup_trainer(callbacks=[pl.callbacks.ModelCheckpoint(dirpath="gene_removal_test_checkpoint")], max_epochs=2)
    model.train()

    checkpoint_path = os.path.join("gene_removal_test_checkpoint", os.listdir("gene_removal_test_checkpoint")[0])
    model = SIMS(weights_path=checkpoint_path)

    predictions = model.predict(testdata)

    shutil.rmtree(os.path.dirname(checkpoint_path))
    assert len(predictions) == 100
    
def test_inference_less_genes_than_train():
    traindata = blobs(n_variables=20, n_observations=100)
    testdata = blobs(n_variables=10, n_observations=100)
    
    traindata.obs["blobs"] = traindata.obs["blobs"].apply(lambda x: f"label_{x}")
    # make data 32-bit
    traindata.X = traindata.X.astype(np.float32)
    testdata.X = testdata.X.astype(np.float32)

    model = SIMS(data=traindata, class_label="blobs")
    model.setup_trainer(callbacks=[pl.callbacks.ModelCheckpoint(dirpath="zero_inflation_test_checkpoint")], max_epochs=2)
    model.train()

    checkpoint_path = os.path.join("zero_inflation_test_checkpoint", os.listdir("zero_inflation_test_checkpoint")[0])
    model = SIMS(weights_path=checkpoint_path)
    predictions = model.predict(testdata, num_workers=0)
    shutil.rmtree(os.path.dirname(checkpoint_path))

    assert len(predictions) == 100

def test_inference_decoded():
    traindata = blobs(n_variables=8, n_observations=100)  # has labels 0,..7 as str
    testdata = blobs(n_variables=10, n_observations=100)
    traindata.obs["blobs"] = traindata.obs["blobs"].apply(lambda x: f"label_{x}")
    traindata.X = traindata.X.astype(np.float32)
    testdata.X = testdata.X.astype(np.float32)

    model = SIMS(data=traindata, class_label="blobs")
    model.setup_trainer(callbacks=[pl.callbacks.ModelCheckpoint(dirpath="test_checkpoints")], max_epochs=2)
    model.train()

    # load in model to check if caching works
    checkpoint_path = os.path.join("test_checkpoints", os.listdir("test_checkpoints")[0])
    model = SIMS(weights_path=checkpoint_path)
    predictions = model.predict(testdata, num_workers=0)

    predicted_labels = predictions["first_pred"].unique()
    # remove test checkpoint
    shutil.rmtree(os.path.dirname(checkpoint_path))
    assert all(x.startswith("label_") for x in predicted_labels)

if __name__ == "__main__":
    test_inference_more_genes_than_train()
    test_inference_less_genes_than_train()
    test_inference_decoded()